@@ -5,6 +5,8 @@
 #include <stdlib.h>
 #include <unistd.h>
 #include <wlr/config.h>
+#include <wlr/types/wlr_data_device.h>
+#include <wlr/types/wlr_primary_selection.h>
 #include <wlr/types/wlr_surface.h>
 #include <wlr/util/edges.h>
 #include <wlr/util/log.h>
@@ -12,11 +14,13 @@
 #include <wlr/xwayland.h>
 #include <xcb/composite.h>
 #include <xcb/render.h>
+#include <xcb/res.h>
+#include <xcb/xcb_icccm.h>
 #include <xcb/xfixes.h>
 #include "util/signal.h"
 #include "xwayland/xwm.h"
 
-const char *atom_map[ATOM_LAST] = {
+const char *const atom_map[ATOM_LAST] = {
 	[WL_SURFACE_ID] = "WL_SURFACE_ID",
 	[WM_DELETE_WINDOW] = "WM_DELETE_WINDOW",
 	[WM_PROTOCOLS] = "WM_PROTOCOLS",
@@ -56,6 +60,9 @@ const char *atom_map[ATOM_LAST] = {
 	[TEXT] = "TEXT",
 	[TIMESTAMP] = "TIMESTAMP",
 	[DELETE] = "DELETE",
+	[NET_STARTUP_ID] = "_NET_STARTUP_ID",
+	[NET_STARTUP_INFO] = "_NET_STARTUP_INFO",
+	[NET_STARTUP_INFO_BEGIN] = "_NET_STARTUP_INFO_BEGIN",
 	[NET_WM_WINDOW_TYPE_NORMAL] = "_NET_WM_WINDOW_TYPE_NORMAL",
 	[NET_WM_WINDOW_TYPE_UTILITY] = "_NET_WM_WINDOW_TYPE_UTILITY",
 	[NET_WM_WINDOW_TYPE_TOOLTIP] = "_NET_WM_WINDOW_TYPE_TOOLTIP",
@@ -81,6 +88,15 @@ const char *atom_map[ATOM_LAST] = {
 	[DND_ACTION_ASK] = "XdndActionAsk",
 	[DND_ACTION_PRIVATE] = "XdndActionPrivate",
 	[NET_CLIENT_LIST] = "_NET_CLIENT_LIST",
+	[NET_CLIENT_LIST_STACKING] = "_NET_CLIENT_LIST_STACKING",
+};
+
+#define STARTUP_INFO_REMOVE_PREFIX "remove: ID="
+struct pending_startup_id {
+	char *msg;
+	size_t len;
+	xcb_window_t window;
+	struct wl_list link;
 };
 
 static const struct wlr_surface_role xwayland_surface_role;
@@ -143,6 +159,7 @@ static struct wlr_xwayland_surface *xwayland_surface_create(
 	surface->height = height;
 	surface->override_redirect = override_redirect;
 	wl_list_init(&surface->children);
+	wl_list_init(&surface->stack_link);
 	wl_list_init(&surface->parent_link);
 	wl_signal_init(&surface->events.destroy);
 	wl_signal_init(&surface->events.request_configure);
@@ -159,6 +176,7 @@ static struct wlr_xwayland_surface *xwayland_surface_create(
 	wl_signal_init(&surface->events.set_title);
 	wl_signal_init(&surface->events.set_parent);
 	wl_signal_init(&surface->events.set_pid);
+	wl_signal_init(&surface->events.set_startup_id);
 	wl_signal_init(&surface->events.set_window_type);
 	wl_signal_init(&surface->events.set_hints);
 	wl_signal_init(&surface->events.set_decorations);
@@ -219,6 +237,10 @@ static void xwm_send_wm_message(struct wlr_xwayland_surface *surface,
 }
 
 static void xwm_set_net_client_list(struct wlr_xwm *xwm) {
+	// FIXME: _NET_CLIENT_LIST is expected to be ordered by map time, but the
+	// order of surfaces in `xwm->surfaces` is by creation time. The order of
+	// windows _NET_CLIENT_LIST exposed by wlroots is wrong.
+
 	size_t mapped_surfaces = 0;
 	struct wlr_xwayland_surface *surface;
 	wl_list_for_each(surface, &xwm->surfaces, link) {
@@ -227,7 +249,11 @@ static void xwm_set_net_client_list(struct wlr_xwm *xwm) {
 		}
 	}
 
-	xcb_window_t windows[mapped_surfaces + 1];
+	xcb_window_t *windows = malloc(sizeof(xcb_window_t) * mapped_surfaces);
+	if (!windows) {
+		return;
+	}
+
 	size_t index = 0;
 	wl_list_for_each(surface, &xwm->surfaces, link) {
 		if (surface->mapped) {
@@ -238,6 +264,29 @@ static void xwm_set_net_client_list(struct wlr_xwm *xwm) {
 	xcb_change_property(xwm->xcb_conn, XCB_PROP_MODE_REPLACE,
 			xwm->screen->root, xwm->atoms[NET_CLIENT_LIST],
 			XCB_ATOM_WINDOW, 32, mapped_surfaces, windows);
+	free(windows);
+}
+
+static void xwm_set_net_client_list_stacking(struct wlr_xwm *xwm) {
+	size_t num_surfaces = wl_list_length(&xwm->surfaces_in_stack_order);
+	xcb_window_t *windows = malloc(sizeof(xcb_window_t) * num_surfaces);
+	if (!windows) {
+		return;
+	}
+
+	// We store surfaces in top-to-bottom order because this is easier to reason
+	// about, but _NET_CLIENT_LIST_STACKING is supposed to be in bottom-to-top
+	// order, so iterate backwards through the list.
+	size_t i = 0;
+	struct wlr_xwayland_surface *xsurface;
+	wl_list_for_each(xsurface, &xwm->surfaces_in_stack_order, stack_link) {
+		windows[i++] = xsurface->window_id;
+	}
+
+	xcb_change_property(xwm->xcb_conn, XCB_PROP_MODE_REPLACE, xwm->screen->root,
+			xwm->atoms[NET_CLIENT_LIST_STACKING], XCB_ATOM_WINDOW, 32, num_surfaces,
+			windows);
+	free(windows);
 }
 
 static void xsurface_set_net_wm_state(struct wlr_xwayland_surface *xsurface);
@@ -281,11 +330,6 @@ static void xwm_set_focus_window(struct wlr_xwm *xwm,
 		xwm->last_focus_seq = cookie.sequence;
 	}
 
-	uint32_t values[1];
-	values[0] = XCB_STACK_MODE_ABOVE;
-	xcb_configure_window(xwm->xcb_conn, xsurface->window_id,
-		XCB_CONFIG_WINDOW_STACK_MODE, values);
-
 	xsurface_set_net_wm_state(xsurface);
 }
 
@@ -354,6 +398,7 @@ static void xwayland_surface_destroy(
 	}
 
 	wl_list_remove(&xsurface->link);
+	wl_list_remove(&xsurface->stack_link);
 	wl_list_remove(&xsurface->parent_link);
 
 	struct wlr_xwayland_surface *child, *next;
@@ -380,6 +425,7 @@ static void xwayland_surface_destroy(
 	free(xsurface->role);
 	free(xsurface->window_type);
 	free(xsurface->protocols);
+	free(xsurface->startup_id);
 	free(xsurface->hints);
 	free(xsurface->size_hints);
 	free(xsurface);
@@ -411,12 +457,32 @@ static void read_surface_class(struct wlr_xwm *xwm,
 		surface->class = NULL;
 	}
 
-	wlr_log(WLR_DEBUG, "XCB_ATOM_WM_CLASS: %s %s",
-		surface->instance ? surface->instance : "(null)",
-		surface->class ? surface->class : "(null)");
 	wlr_signal_emit_safe(&surface->events.set_class, surface);
 }
 
+static void read_surface_startup_id(struct wlr_xwm *xwm,
+		struct wlr_xwayland_surface *xsurface,
+		xcb_get_property_reply_t *reply) {
+	if (reply->type != XCB_ATOM_STRING &&
+			reply->type != xwm->atoms[UTF8_STRING]) {
+		return;
+	}
+
+	size_t len = xcb_get_property_value_length(reply);
+	char *startup_id = xcb_get_property_value(reply);
+
+	free(xsurface->startup_id);
+	if (len > 0) {
+		xsurface->startup_id = strndup(startup_id, len);
+	} else {
+		xsurface->startup_id = NULL;
+	}
+
+	wlr_log(WLR_DEBUG, "XCB_ATOM_NET_STARTUP_ID: %s",
+		xsurface->startup_id ? xsurface->startup_id: "(null)");
+	wlr_signal_emit_safe(&xsurface->events.set_startup_id, xsurface);
+}
+
 static void read_surface_role(struct wlr_xwm *xwm,
 		struct wlr_xwayland_surface *xsurface,
 		xcb_get_property_reply_t *reply) {
@@ -435,8 +501,6 @@ static void read_surface_role(struct wlr_xwm *xwm,
 		xsurface->role = NULL;
 	}
 
-	wlr_log(WLR_DEBUG, "XCB_ATOM_WM_WINDOW_ROLE: %s",
-		xsurface->role ? xsurface->role : "(null)");
 	wlr_signal_emit_safe(&xsurface->events.set_role, xsurface);
 }
 
@@ -464,7 +528,6 @@ static void read_surface_title(struct wlr_xwm *xwm,
 	}
 	xsurface->has_utf8_title = is_utf8;
 
-	wlr_log(WLR_DEBUG, "XCB_ATOM_WM_NAME: %s", xsurface->title ? xsurface->title : "(null)");
 	wlr_signal_emit_safe(&xsurface->events.set_title, xsurface);
 }
 
@@ -510,21 +573,42 @@ static void read_surface_parent(struct wlr_xwm *xwm,
 		wl_list_init(&xsurface->parent_link);
 	}
 
-	wlr_log(WLR_DEBUG, "XCB_ATOM_WM_TRANSIENT_FOR: %p", xsurface->parent);
 	wlr_signal_emit_safe(&xsurface->events.set_parent, xsurface);
 }
 
-static void read_surface_pid(struct wlr_xwm *xwm,
-		struct wlr_xwayland_surface *xsurface,
-		xcb_get_property_reply_t *reply) {
-	if (reply->type != XCB_ATOM_CARDINAL) {
+static void read_surface_client_id(struct wlr_xwm *xwm,
+		struct wlr_xwayland_surface *xsurface) {
+	xcb_res_client_id_spec_t spec = {
+		.client = xsurface->window_id,
+		.mask = XCB_RES_CLIENT_ID_MASK_LOCAL_CLIENT_PID
+	};
+
+	xcb_res_query_client_ids_cookie_t cookie = xcb_res_query_client_ids(
+		xwm->xcb_conn, 1, &spec);
+	xcb_res_query_client_ids_reply_t *reply = xcb_res_query_client_ids_reply(
+		xwm->xcb_conn, cookie,  NULL);
+	if (reply == NULL) {
 		return;
 	}
 
-	pid_t *pid = xcb_get_property_value(reply);
+	uint32_t *pid = NULL;
+	xcb_res_client_id_value_iterator_t iter =
+		xcb_res_query_client_ids_ids_iterator(reply);
+	while (iter.rem > 0) {
+		if (iter.data->spec.mask & XCB_RES_CLIENT_ID_MASK_LOCAL_CLIENT_PID &&
+				xcb_res_client_id_value_value_length(iter.data) > 0) {
+			pid = xcb_res_client_id_value_value(iter.data);
+			break;
+		}
+		xcb_res_client_id_value_next(&iter);
+	}
+	if (pid == NULL) {
+		free(reply);
+		return;
+	}
 	xsurface->pid = *pid;
-	wlr_log(WLR_DEBUG, "NET_WM_PID %d", xsurface->pid);
 	wlr_signal_emit_safe(&xsurface->events.set_pid, xsurface);
+	free(reply);
 }
 
 static void read_surface_window_type(struct wlr_xwm *xwm,
@@ -546,7 +630,6 @@ static void read_surface_window_type(struct wlr_xwm *xwm,
 	memcpy(xsurface->window_type, atoms, atoms_size);
 	xsurface->window_type_len = atoms_len;
 
-	wlr_log(WLR_DEBUG, "NET_WM_WINDOW_TYPE (%zu)", atoms_len);
 	wlr_signal_emit_safe(&xsurface->events.set_window_type, xsurface);
 }
 
@@ -568,11 +651,8 @@ static void read_surface_protocols(struct wlr_xwm *xwm,
 	}
 	memcpy(xsurface->protocols, atoms, atoms_size);
 	xsurface->protocols_len = atoms_len;
-
-	wlr_log(WLR_DEBUG, "WM_PROTOCOLS (%zu)", atoms_len);
 }
 
-#if WLR_HAS_XCB_ICCCM
 static void read_surface_hints(struct wlr_xwm *xwm,
 		struct wlr_xwayland_surface *xsurface,
 		xcb_get_property_reply_t *reply) {
@@ -590,6 +670,7 @@ static void read_surface_hints(struct wlr_xwm *xwm,
 	if (xsurface->hints == NULL) {
 		return;
 	}
+
 	memcpy(xsurface->hints, &hints, sizeof(struct wlr_xwayland_surface_hints));
 	xsurface->hints_urgency = xcb_icccm_wm_hints_get_urgency(&hints);
 
@@ -599,18 +680,9 @@ static void read_surface_hints(struct wlr_xwm *xwm,
 		xsurface->hints->input = true;
 	}
 
-	wlr_log(WLR_DEBUG, "WM_HINTS (%" PRIu32 ")", reply->value_len);
 	wlr_signal_emit_safe(&xsurface->events.set_hints, xsurface);
 }
-#else
-static void read_surface_hints(struct wlr_xwm *xwm,
-		struct wlr_xwayland_surface *xsurface,
-		xcb_get_property_reply_t *reply) {
-	// Do nothing
-}
-#endif
 
-#if WLR_HAS_XCB_ICCCM
 static void read_surface_normal_hints(struct wlr_xwm *xwm,
 		struct wlr_xwayland_surface *xsurface,
 		xcb_get_property_reply_t *reply) {
@@ -650,17 +722,7 @@ static void read_surface_normal_hints(struct wlr_xwm *xwm,
 		xsurface->size_hints->max_width = -1;
 		xsurface->size_hints->max_height = -1;
 	}
-
-	wlr_log(WLR_DEBUG, "WM_NORMAL_HINTS (%" PRIu32 ")", reply->value_len);
 }
-#else
-static void read_surface_normal_hints(struct wlr_xwm *xwm,
-		struct wlr_xwayland_surface *xsurface,
-		xcb_get_property_reply_t *reply) {
-	// Do nothing
-}
-#endif
-
 
 #define MWM_HINTS_FLAGS_FIELD 0
 #define MWM_HINTS_DECORATIONS_FIELD 2
@@ -694,8 +756,6 @@ static void read_surface_motif_hints(struct wlr_xwm *xwm,
 		}
 		wlr_signal_emit_safe(&xsurface->events.set_decorations, xsurface);
 	}
-
-	wlr_log(WLR_DEBUG, "MOTIF_WM_HINTS (%" PRIu32 ")", reply->value_len);
 }
 
 static void read_surface_net_wm_state(struct wlr_xwm *xwm,
@@ -751,7 +811,7 @@ static void read_surface_property(struct wlr_xwm *xwm,
 	} else if (property == XCB_ATOM_WM_TRANSIENT_FOR) {
 		read_surface_parent(xwm, xsurface, reply);
 	} else if (property == xwm->atoms[NET_WM_PID]) {
-		read_surface_pid(xwm, xsurface, reply);
+		// intentionally ignored
 	} else if (property == xwm->atoms[NET_WM_WINDOW_TYPE]) {
 		read_surface_window_type(xwm, xsurface, reply);
 	} else if (property == xwm->atoms[WM_PROTOCOLS]) {
@@ -766,6 +826,8 @@ static void read_surface_property(struct wlr_xwm *xwm,
 		read_surface_motif_hints(xwm, xsurface, reply);
 	} else if (property == xwm->atoms[WM_WINDOW_ROLE]) {
 		read_surface_role(xwm, xsurface, reply);
+	} else if (property == xwm->atoms[NET_STARTUP_ID]) {
+		read_surface_startup_id(xwm, xsurface, reply);
 	} else {
 		char *prop_name = xwm_get_atom_name(xwm, property);
 		wlr_log(WLR_DEBUG, "unhandled X11 property %" PRIu32 " (%s) for window %" PRIu32,
@@ -798,7 +860,7 @@ static void xwayland_surface_role_precommit(struct wlr_surface *wlr_surface) {
 	}
 
 	if (wlr_surface->pending.committed & WLR_SURFACE_STATE_BUFFER &&
-			wlr_surface->pending.buffer_resource == NULL) {
+			wlr_surface->pending.buffer == NULL) {
 		// This is a NULL commit
 		if (surface->mapped) {
 			wlr_signal_emit_safe(&surface->events.unmap, surface);
@@ -839,14 +901,17 @@ static void xwm_map_shell_surface(struct wlr_xwm *xwm,
 		xwm->atoms[WM_HINTS],
 		xwm->atoms[WM_NORMAL_HINTS],
 		xwm->atoms[MOTIF_WM_HINTS],
+		xwm->atoms[NET_STARTUP_ID],
 		xwm->atoms[NET_WM_STATE],
 		xwm->atoms[NET_WM_WINDOW_TYPE],
 		xwm->atoms[NET_WM_NAME],
-		xwm->atoms[NET_WM_PID],
 	};
 	for (size_t i = 0; i < sizeof(props)/sizeof(xcb_atom_t); i++) {
 		read_surface_property(xwm, xsurface, props[i]);
 	}
+	if (xwm->xres) {
+		read_surface_client_id(xwm, xsurface);
+	}
 
 	xsurface->surface_destroy.notify = handle_surface_destroy;
 	wl_signal_add(&surface->events.destroy, &xsurface->surface_destroy);
@@ -876,11 +941,10 @@ static void xsurface_unmap(struct wlr_xwayland_surface *surface) {
 
 static void xwm_handle_create_notify(struct wlr_xwm *xwm,
 		xcb_create_notify_event_t *ev) {
-	wlr_log(WLR_DEBUG, "XCB_CREATE_NOTIFY (%u)", ev->window);
-
 	if (ev->window == xwm->window ||
-			ev->window == xwm->selection_window ||
-			ev->window == xwm->dnd_window) {
+			ev->window == xwm->primary_selection.window ||
+			ev->window == xwm->clipboard_selection.window ||
+			ev->window == xwm->dnd_selection.window) {
 		return;
 	}
 
@@ -890,7 +954,6 @@ static void xwm_handle_create_notify(struct wlr_xwm *xwm,
 
 static void xwm_handle_destroy_notify(struct wlr_xwm *xwm,
 		xcb_destroy_notify_event_t *ev) {
-	wlr_log(WLR_DEBUG, "XCB_DESTROY_NOTIFY (%u)", ev->window);
 	struct wlr_xwayland_surface *xsurface = lookup_surface(xwm, ev->window);
 	if (xsurface == NULL) {
 		return;
@@ -923,8 +986,6 @@ static void xwm_handle_configure_request(struct wlr_xwm *xwm,
 		.height = mask & XCB_CONFIG_WINDOW_HEIGHT ? ev->height : surface->height,
 		.mask = mask,
 	};
-	wlr_log(WLR_DEBUG, "XCB_CONFIGURE_REQUEST (%u) [%ux%u+%d,%d]", ev->window,
-		wlr_event.width, wlr_event.height, wlr_event.x, wlr_event.y);
 
 	wlr_signal_emit_safe(&surface->events.request_configure, &wlr_event);
 }
@@ -957,10 +1018,6 @@ static void xwm_handle_configure_notify(struct wlr_xwm *xwm,
 	}
 }
 
-#define ICCCM_WITHDRAWN_STATE	0
-#define ICCCM_NORMAL_STATE	1
-#define ICCCM_ICONIC_STATE	3
-
 static void xsurface_set_wm_state(struct wlr_xwayland_surface *xsurface,
 		int32_t state) {
 	struct wlr_xwm *xwm = xsurface->xwm;
@@ -975,28 +1032,63 @@ static void xsurface_set_wm_state(struct wlr_xwayland_surface *xsurface,
 		sizeof(property) / sizeof(property[0]), property);
 }
 
+void wlr_xwayland_surface_restack(struct wlr_xwayland_surface *xsurface,
+		struct wlr_xwayland_surface *sibling, enum xcb_stack_mode_t mode) {
+	struct wlr_xwm *xwm = xsurface->xwm;
+	uint32_t values[2];
+	size_t idx = 0;
+	uint32_t flags = XCB_CONFIG_WINDOW_STACK_MODE;
+
+	if (sibling != NULL) {
+		values[idx++] = sibling->window_id;
+		flags |= XCB_CONFIG_WINDOW_SIBLING;
+	}
+	values[idx++] = mode;
+
+	xcb_configure_window(xwm->xcb_conn, xsurface->window_id, flags, values);
+
+	wl_list_remove(&xsurface->stack_link);
+
+	struct wl_list *node;
+	if (mode == XCB_STACK_MODE_ABOVE) {
+		if (sibling) {
+			node = &sibling->stack_link;
+		} else {
+			node = xwm->surfaces_in_stack_order.prev;
+		}
+	} else if (mode == XCB_STACK_MODE_BELOW) {
+		if (sibling) {
+			node = sibling->stack_link.prev;
+		} else {
+			node = &xwm->surfaces_in_stack_order;
+		}
+	} else {
+		// Not implementing XCB_STACK_MODE_TOP_IF | XCB_STACK_MODE_BOTTOM_IF |
+		// XCB_STACK_MODE_OPPOSITE.
+		abort();
+	}
+
+	wl_list_insert(node, &xsurface->stack_link);
+	xwm_set_net_client_list_stacking(xwm);
+	xcb_flush(xwm->xcb_conn);
+}
+
 static void xwm_handle_map_request(struct wlr_xwm *xwm,
 		xcb_map_request_event_t *ev) {
-	wlr_log(WLR_DEBUG, "XCB_MAP_REQUEST (%u)", ev->window);
 	struct wlr_xwayland_surface *xsurface = lookup_surface(xwm, ev->window);
 	if (!xsurface) {
 		return;
 	}
 
-	xsurface_set_wm_state(xsurface, ICCCM_NORMAL_STATE);
+	xsurface_set_wm_state(xsurface, XCB_ICCCM_WM_STATE_NORMAL);
 	xsurface_set_net_wm_state(xsurface);
 
-	uint32_t values[1];
-	values[0] = XCB_STACK_MODE_BELOW;
-	xcb_configure_window(xwm->xcb_conn, ev->window,
-			XCB_CONFIG_WINDOW_STACK_MODE, values);
-
+	wlr_xwayland_surface_restack(xsurface, NULL, XCB_STACK_MODE_BELOW);
 	xcb_map_window(xwm->xcb_conn, ev->window);
 }
 
 static void xwm_handle_map_notify(struct wlr_xwm *xwm,
 		xcb_map_notify_event_t *ev) {
-	wlr_log(WLR_DEBUG, "XCB_MAP_NOTIFY (%u)", ev->window);
 	struct wlr_xwayland_surface *xsurface = lookup_surface(xwm, ev->window);
 	if (!xsurface) {
 		return;
@@ -1010,19 +1102,17 @@ static void xwm_handle_map_notify(struct wlr_xwm *xwm,
 
 static void xwm_handle_unmap_notify(struct wlr_xwm *xwm,
 		xcb_unmap_notify_event_t *ev) {
-	wlr_log(WLR_DEBUG, "XCB_UNMAP_NOTIFY (%u)", ev->window);
 	struct wlr_xwayland_surface *xsurface = lookup_surface(xwm, ev->window);
 	if (xsurface == NULL) {
 		return;
 	}
 
 	xsurface_unmap(xsurface);
-	xsurface_set_wm_state(xsurface, ICCCM_WITHDRAWN_STATE);
+	xsurface_set_wm_state(xsurface, XCB_ICCCM_WM_STATE_WITHDRAWN);
 }
 
 static void xwm_handle_property_notify(struct wlr_xwm *xwm,
 		xcb_property_notify_event_t *ev) {
-	wlr_log(WLR_DEBUG, "XCB_PROPERTY_NOTIFY (%u)", ev->window);
 	struct wlr_xwayland_surface *xsurface = lookup_surface(xwm, ev->window);
 	if (xsurface == NULL) {
 		return;
@@ -1165,7 +1255,7 @@ static bool update_state(int action, bool *state) {
 	return changed;
 }
 
-static inline bool xsurface_is_maximized(
+static bool xsurface_is_maximized(
 		struct wlr_xwayland_surface *xsurface) {
 	return xsurface->maximized_horz && xsurface->maximized_vert;
 }
@@ -1189,27 +1279,27 @@ static void xwm_handle_net_wm_state_message(struct wlr_xwm *xwm,
 	for (size_t i = 0; i < 2; ++i) {
 		xcb_atom_t property = client_message->data.data32[1 + i];
 
-		if (property == xwm->atoms[NET_WM_STATE_MODAL] &&
-				update_state(action, &xsurface->modal)) {
-			xsurface_set_net_wm_state(xsurface);
-		} else if (property == xwm->atoms[NET_WM_STATE_FULLSCREEN] &&
-				update_state(action, &xsurface->fullscreen)) {
-			xsurface_set_net_wm_state(xsurface);
-		} else if (property == xwm->atoms[NET_WM_STATE_MAXIMIZED_VERT] &&
-				update_state(action, &xsurface->maximized_vert)) {
-			xsurface_set_net_wm_state(xsurface);
-		} else if (property == xwm->atoms[NET_WM_STATE_MAXIMIZED_HORZ] &&
-				update_state(action, &xsurface->maximized_horz)) {
-			xsurface_set_net_wm_state(xsurface);
-		} else if (property == xwm->atoms[NET_WM_STATE_HIDDEN] &&
-				update_state(action, &xsurface->minimized)) {
-			xsurface_set_net_wm_state(xsurface);
+		bool changed = false;
+		if (property == xwm->atoms[NET_WM_STATE_MODAL]) {
+			changed = update_state(action, &xsurface->modal);
+		} else if (property == xwm->atoms[NET_WM_STATE_FULLSCREEN]) {
+			changed = update_state(action, &xsurface->fullscreen);
+		} else if (property == xwm->atoms[NET_WM_STATE_MAXIMIZED_VERT]) {
+			changed = update_state(action, &xsurface->maximized_vert);
+		} else if (property == xwm->atoms[NET_WM_STATE_MAXIMIZED_HORZ]) {
+			changed = update_state(action, &xsurface->maximized_horz);
+		} else if (property == xwm->atoms[NET_WM_STATE_HIDDEN]) {
+			changed = update_state(action, &xsurface->minimized);
 		} else if (property != XCB_ATOM_NONE) {
 			char *prop_name = xwm_get_atom_name(xwm, property);
 			wlr_log(WLR_DEBUG, "Unhandled NET_WM_STATE property change "
 				"%"PRIu32" (%s)", property, prop_name ? prop_name : "(null)");
 			free(prop_name);
 		}
+
+		if (changed) {
+			xsurface_set_net_wm_state(xsurface);
+		}
 	}
 	// client_message->data.data32[3] is the source indication
 	// all other values are set to 0
@@ -1281,6 +1371,73 @@ static void xwm_handle_net_active_window_message(struct wlr_xwm *xwm,
 	wlr_signal_emit_safe(&surface->events.request_activate, surface);
 }
 
+static void pending_startup_id_destroy(struct pending_startup_id *pending) {
+	wl_list_remove(&pending->link);
+	free(pending->msg);
+	free(pending);
+}
+
+static void xwm_handle_net_startup_info_message(struct wlr_xwm *xwm,
+		xcb_client_message_event_t *ev) {
+	struct pending_startup_id *pending, *curr = NULL;
+	wl_list_for_each(pending, &xwm->pending_startup_ids, link) {
+		if (pending->window == ev->window) {
+			curr = pending;
+			break;
+		}
+	}
+
+	char *start;
+	size_t buf_len = sizeof(ev->data);
+	if (curr) {
+		curr->msg = realloc(curr->msg, curr->len + buf_len);
+		if (!curr->msg) {
+			pending_startup_id_destroy(curr);
+			return;
+		}
+		start = curr->msg + curr->len;
+		curr->len += buf_len;
+	} else {
+		curr = calloc(1, sizeof(struct pending_startup_id));
+		if (!curr)
+			return;
+		curr->window = ev->window;
+		curr->msg = malloc(buf_len);
+		if (!curr->msg) {
+			free(curr);
+			return;
+		}
+		start = curr->msg;
+		curr->len = buf_len;
+		wl_list_insert(&xwm->pending_startup_ids, &curr->link);
+	}
+
+	char *id = NULL;
+	const char *data = (const char *)ev->data.data8;
+	for (size_t i = 0; i < buf_len; i++) {
+		start[i] = data[i];
+		if (start[i] == '\0') {
+			if (strncmp(curr->msg, STARTUP_INFO_REMOVE_PREFIX,
+					strlen(STARTUP_INFO_REMOVE_PREFIX)) == 0 &&
+					strlen(curr->msg) > strlen(STARTUP_INFO_REMOVE_PREFIX)) {
+				id = curr->msg + strlen(STARTUP_INFO_REMOVE_PREFIX);
+				break;
+			} else {
+				wlr_log(WLR_ERROR, "Unhandled message '%s'\n", curr->msg);
+				pending_startup_id_destroy(curr);
+				return;
+			}
+		}
+	}
+
+	if (id) {
+		struct wlr_xwayland_remove_startup_info_event data = { id, ev->window };
+		wlr_log(WLR_DEBUG, "Got startup id: %s", id);
+		wlr_signal_emit_safe(&xwm->xwayland->events.remove_startup_info, &data);
+		pending_startup_id_destroy(curr);
+	}
+}
+
 static void xwm_handle_wm_change_state_message(struct wlr_xwm *xwm,
 		xcb_client_message_event_t *ev) {
 	struct wlr_xwayland_surface *xsurface = lookup_surface(xwm, ev->window);
@@ -1291,9 +1448,9 @@ static void xwm_handle_wm_change_state_message(struct wlr_xwm *xwm,
 	}
 
 	bool minimize;
-	if (detail == ICCCM_ICONIC_STATE) {
+	if (detail == XCB_ICCCM_WM_STATE_ICONIC) {
 		minimize = true;
-	} else if (detail == ICCCM_NORMAL_STATE) {
+	} else if (detail == XCB_ICCCM_WM_STATE_NORMAL) {
 		minimize = false;
 	} else {
 		wlr_log(WLR_DEBUG, "unhandled wm_change_state event %u", detail);
@@ -1309,8 +1466,6 @@ static void xwm_handle_wm_change_state_message(struct wlr_xwm *xwm,
 
 static void xwm_handle_client_message(struct wlr_xwm *xwm,
 		xcb_client_message_event_t *ev) {
-	wlr_log(WLR_DEBUG, "XCB_CLIENT_MESSAGE (%u)", ev->window);
-
 	if (ev->type == xwm->atoms[WL_SURFACE_ID]) {
 		xwm_handle_surface_id_message(xwm, ev);
 	} else if (ev->type == xwm->atoms[NET_WM_STATE]) {
@@ -1321,6 +1476,9 @@ static void xwm_handle_client_message(struct wlr_xwm *xwm,
 		xwm_handle_wm_protocols_message(xwm, ev);
 	} else if (ev->type == xwm->atoms[NET_ACTIVE_WINDOW]) {
 		xwm_handle_net_active_window_message(xwm, ev);
+	} else if (ev->type == xwm->atoms[NET_STARTUP_INFO] ||
+			ev->type == xwm->atoms[NET_STARTUP_INFO_BEGIN]) {
+		xwm_handle_net_startup_info_message(xwm, ev);
 	} else if (ev->type == xwm->atoms[WM_CHANGE_STATE]) {
 		xwm_handle_wm_change_state_message(xwm, ev);
 	} else if (!xwm_handle_selection_client_message(xwm, ev)) {
@@ -1373,7 +1531,7 @@ static void xwm_handle_focus_in(struct wlr_xwm *xwm,
 }
 
 static void xwm_handle_xcb_error(struct wlr_xwm *xwm, xcb_value_error_t *ev) {
-#if WLR_HAS_XCB_ERRORS
+#if HAS_XCB_ERRORS
 	const char *major_name =
 		xcb_errors_get_name_for_major_code(xwm->errors_context,
 			ev->major_opcode);
@@ -1411,7 +1569,7 @@ log_raw:
 }
 
 static void xwm_handle_unhandled_event(struct wlr_xwm *xwm, xcb_generic_event_t *ev) {
-#if WLR_HAS_XCB_ERRORS
+#if HAS_XCB_ERRORS
 	const char *extension;
 	const char *event_name =
 		xcb_errors_get_name_for_xcb_event(xwm->errors_context,
@@ -1590,7 +1748,28 @@ void xwm_destroy(struct wlr_xwm *xwm) {
 	if (!xwm) {
 		return;
 	}
-	xwm_selection_finish(xwm);
+
+	xwm_selection_finish(&xwm->clipboard_selection);
+	xwm_selection_finish(&xwm->primary_selection);
+	xwm_selection_finish(&xwm->dnd_selection);
+
+	if (xwm->seat) {
+		if (xwm->seat->selection_source &&
+				data_source_is_xwayland(xwm->seat->selection_source)) {
+			wlr_seat_set_selection(xwm->seat, NULL,
+				wl_display_next_serial(xwm->xwayland->wl_display));
+		}
+
+		if (xwm->seat->primary_selection_source &&
+				primary_selection_source_is_xwayland(
+					xwm->seat->primary_selection_source)) {
+			wlr_seat_set_primary_selection(xwm->seat, NULL,
+				wl_display_next_serial(xwm->xwayland->wl_display));
+		}
+
+		wlr_xwayland_set_seat(xwm->xwayland, NULL);
+	}
+
 	if (xwm->cursor) {
 		xcb_free_cursor(xwm->xcb_conn, xwm->cursor);
 	}
@@ -1603,7 +1782,7 @@ void xwm_destroy(struct wlr_xwm *xwm) {
 	if (xwm->event_source) {
 		wl_event_source_remove(xwm->event_source);
 	}
-#if WLR_HAS_XCB_ERRORS
+#if HAS_XCB_ERRORS
 	if (xwm->errors_context) {
 		xcb_errors_context_free(xwm->errors_context);
 	}
@@ -1612,13 +1791,18 @@ void xwm_destroy(struct wlr_xwm *xwm) {
 	wl_list_for_each_safe(xsurface, tmp, &xwm->surfaces, link) {
 		xwayland_surface_destroy(xsurface);
 	}
-	wl_list_for_each_safe(xsurface, tmp, &xwm->unpaired_surfaces, link) {
+	wl_list_for_each_safe(xsurface, tmp, &xwm->unpaired_surfaces, unpaired_link) {
 		xwayland_surface_destroy(xsurface);
 	}
 	wl_list_remove(&xwm->compositor_new_surface.link);
 	wl_list_remove(&xwm->compositor_destroy.link);
 	xcb_disconnect(xwm->xcb_conn);
 
+	struct pending_startup_id *pending, *next;
+	wl_list_for_each_safe(pending, next, &xwm->pending_startup_ids, link) {
+		pending_startup_id_destroy(pending);
+	}
+
 	xwm->xwayland->xwm = NULL;
 	free(xwm);
 }
@@ -1626,6 +1810,7 @@ void xwm_destroy(struct wlr_xwm *xwm) {
 static void xwm_get_resources(struct wlr_xwm *xwm) {
 	xcb_prefetch_extension_data(xwm->xcb_conn, &xcb_xfixes_id);
 	xcb_prefetch_extension_data(xwm->xcb_conn, &xcb_composite_id);
+	xcb_prefetch_extension_data(xwm->xcb_conn, &xcb_res_id);
 
 	size_t i;
 	xcb_intern_atom_cookie_t cookies[ATOM_LAST];
@@ -1669,6 +1854,29 @@ static void xwm_get_resources(struct wlr_xwm *xwm) {
 		xfixes_reply->major_version, xfixes_reply->minor_version);
 
 	free(xfixes_reply);
+
+	const xcb_query_extension_reply_t *xres =
+		xcb_get_extension_data(xwm->xcb_conn, &xcb_res_id);
+	if (!xres || !xres->present) {
+		return;
+	}
+
+	xcb_res_query_version_cookie_t xres_cookie =
+		xcb_res_query_version(xwm->xcb_conn, XCB_RES_MAJOR_VERSION,
+			XCB_RES_MINOR_VERSION);
+	xcb_res_query_version_reply_t *xres_reply =
+		xcb_res_query_version_reply(xwm->xcb_conn, xres_cookie, NULL);
+	if (xres_reply == NULL) {
+		return;
+	}
+
+	wlr_log(WLR_DEBUG, "xres version: %" PRIu32 ".%" PRIu32,
+		xres_reply->server_major, xres_reply->server_minor);
+	if (xres_reply->server_major > 1 ||
+			(xres_reply->server_major == 1 && xres_reply->server_minor >= 2)) {
+		xwm->xres = xres;
+	}
+	free(xres_reply);
 }
 
 static void xwm_create_wm_window(struct wlr_xwm *xwm) {
@@ -1817,6 +2025,7 @@ void xwm_set_cursor(struct wlr_xwm *xwm, const uint8_t *pixels, uint32_t stride,
 	xcb_render_create_cursor(xwm->xcb_conn, xwm->cursor, pic, hotspot_x,
 		hotspot_y);
 	xcb_free_pixmap(xwm->xcb_conn, pix);
+	xcb_render_free_picture(xwm->xcb_conn, pic);
 
 	uint32_t values[] = {xwm->cursor};
 	xcb_change_window_attributes(xwm->xcb_conn, xwm->screen->root,
@@ -1832,7 +2041,9 @@ struct wlr_xwm *xwm_create(struct wlr_xwayland *xwayland, int wm_fd) {
 
 	xwm->xwayland = xwayland;
 	wl_list_init(&xwm->surfaces);
+	wl_list_init(&xwm->surfaces_in_stack_order);
 	wl_list_init(&xwm->unpaired_surfaces);
+	wl_list_init(&xwm->pending_startup_ids);
 	xwm->ping_timeout = 10000;
 
 	xwm->xcb_conn = xcb_connect_to_fd(wm_fd, NULL);
@@ -1844,7 +2055,7 @@ struct wlr_xwm *xwm_create(struct wlr_xwayland *xwayland, int wm_fd) {
 		return NULL;
 	}
 
-#if WLR_HAS_XCB_ERRORS
+#if HAS_XCB_ERRORS
 	if (xcb_errors_context_new(xwm->xcb_conn, &xwm->errors_context)) {
 		wlr_log(WLR_ERROR, "Could not allocate error context");
 		xwm_destroy(xwm);
@@ -1891,6 +2102,7 @@ struct wlr_xwm *xwm_create(struct wlr_xwayland *xwayland, int wm_fd) {
 		xwm->atoms[NET_WM_STATE_MAXIMIZED_HORZ],
 		xwm->atoms[NET_WM_STATE_HIDDEN],
 		xwm->atoms[NET_CLIENT_LIST],
+		xwm->atoms[NET_CLIENT_LIST_STACKING],
 	};
 	xcb_change_property(xwm->xcb_conn,
 		XCB_PROP_MODE_REPLACE,
@@ -1905,7 +2117,9 @@ struct wlr_xwm *xwm_create(struct wlr_xwayland *xwayland, int wm_fd) {
 
 	xwm_set_net_active_window(xwm, XCB_WINDOW_NONE);
 
-	xwm_selection_init(xwm);
+	xwm_selection_init(&xwm->clipboard_selection, xwm, xwm->atoms[CLIPBOARD]);
+	xwm_selection_init(&xwm->primary_selection, xwm, xwm->atoms[PRIMARY]);
+	xwm_selection_init(&xwm->dnd_selection, xwm, xwm->atoms[DND_SELECTION]);
 
 	xwm->compositor_new_surface.notify = handle_compositor_new_surface;
 	wl_signal_add(&xwayland->compositor->events.new_surface,
@@ -1926,9 +2140,9 @@ void wlr_xwayland_surface_set_minimized(struct wlr_xwayland_surface *surface,
 	surface->minimized = minimized;
 
 	if (minimized) {
-		xsurface_set_wm_state(surface, ICCCM_ICONIC_STATE);
+		xsurface_set_wm_state(surface, XCB_ICCCM_WM_STATE_ICONIC);
 	} else {
-		xsurface_set_wm_state(surface, ICCCM_NORMAL_STATE);
+		xsurface_set_wm_state(surface, XCB_ICCCM_WM_STATE_NORMAL);
 	}
 
 	xsurface_set_net_wm_state(surface);
@@ -1977,8 +2191,7 @@ void wlr_xwayland_surface_ping(struct wlr_xwayland_surface *surface) {
 }
 
 bool wlr_xwayland_or_surface_wants_focus(
-		const struct wlr_xwayland_surface *surface) {
-	bool ret = true;
+		const struct wlr_xwayland_surface *xsurface) {
 	static enum atom_name needles[] = {
 		NET_WM_WINDOW_TYPE_COMBO,
 		NET_WM_WINDOW_TYPE_DND,
@@ -1990,12 +2203,32 @@ bool wlr_xwayland_or_surface_wants_focus(
 		NET_WM_WINDOW_TYPE_TOOLTIP,
 		NET_WM_WINDOW_TYPE_UTILITY,
 	};
+
 	for (size_t i = 0; i < sizeof(needles) / sizeof(needles[0]); ++i) {
-		if (xwm_atoms_contains(surface->xwm, surface->window_type,
-				surface->window_type_len, needles[i])) {
-			ret = false;
+		if (xwm_atoms_contains(xsurface->xwm, xsurface->window_type,
+				xsurface->window_type_len, needles[i])) {
+			return false;
 		}
 	}
 
-	return ret;
+	return true;
+}
+
+enum wlr_xwayland_icccm_input_model wlr_xwayland_icccm_input_model(
+	const struct wlr_xwayland_surface *xsurface) {
+	bool take_focus = xwm_atoms_contains(xsurface->xwm,
+		xsurface->protocols, xsurface->protocols_len,
+		WM_TAKE_FOCUS);
+
+	if (!xsurface->hints || xsurface->hints->input) {
+		if (take_focus) {
+			return WLR_ICCCM_INPUT_MODEL_LOCAL;
+		}
+		return WLR_ICCCM_INPUT_MODEL_PASSIVE;
+	} else {
+		if (take_focus) {
+			return WLR_ICCCM_INPUT_MODEL_GLOBAL;
+		}
+	}
+	return WLR_ICCCM_INPUT_MODEL_NONE;
 }
